import logging

import grpc
import json
import time
import copy
import typing


from deepdiff import DeepDiff
from google.protobuf.json_format import MessageToDict
from perxis.collections import collections_pb2_grpc, collections_pb2
from perxis.roles import roles_pb2_grpc, roles_pb2
from perxis.items import items_pb2_grpc, items_pb2
from perxis.common import common_pb2
from perxis.clients import clients_pb2_grpc, clients_pb2
from perxis.environments import environments_pb2_grpc, environments_pb2
from perxis.extensions.actions import make_action_item, ACTIONS_COLLECTION_ID
from perxis.extensions.item_models import AbstractItem, SyncPolicyItem
from perxis.provider import PerxisItemsWrapper


logger = logging.getLogger(__name__)


class ExtensionSetup:
    def __init__(
            self,
            collections_service: collections_pb2_grpc.CollectionsStub,
            environments_service: environments_pb2_grpc.EnvironmentsStub,
            roles_service: roles_pb2_grpc.RolesStub,
            clients_service: clients_pb2_grpc.ClientsStub,
            items_service: items_pb2_grpc.ItemsStub,
    ):
        self.collections = []
        self.clients = []
        self.roles = []
        self.actions = []
        self.items = []

        self.roles_service = roles_service
        self.clients_service = clients_service
        self.collections_service = collections_service
        self.environments_service = environments_service
        self.items_service = items_service

        self.__max_attempts_count = 5
        self.__sleep_time = 1

    def add_action(self, action: dict):
        self.actions.append(action)

    def set_actions(self, actions: list[dict]):
        self.actions = actions

    def add_collection(self, collection: collections_pb2.Collection):
        self.collections.append(collection)

    def set_collections(self, collections: list[collections_pb2.Collection]):
        self.collections = collections

    def add_role(self, role: roles_pb2.Role):
        self.roles.append(role)

    def set_roles(self, roles: list[roles_pb2.Role]):
        self.roles = roles

    def add_client(self, client: clients_pb2.Client):
        self.clients.append(client)

    def set_clients(self, clients: list[clients_pb2.Client]):
        self.clients = clients

    def set_items(self, items: list[AbstractItem]):
        self.items = items

    # Работа с ролями
    async def __remove_roles(self, space_id: str) -> list[str]:
        errors_list: list[str] = []

        for role in self.roles:
            try:
                await self.roles_service.Delete(
                    roles_pb2.DeleteRequest(
                        space_id=space_id, role_id=role.id
                    )
                )
            except grpc.RpcError as e:
                # Если роли не существует считать это ошибкой не надо
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось удалить роль {role.id}, {e.details()}")

        return errors_list

    async def __check_roles(self, space_id: str) -> list[str]:
        errors_list = []

        for role in self.roles:
            try:
                await self.roles_service.Get(
                    roles_pb2.GetRequest(space_id=space_id, role_id=role.id)
                )
            except grpc.RpcError as e:
                errors_list.append(f"Не удалось получить роль {role.id}, ошибка {e.details()}")

        return errors_list

    async def __update_roles(self, space_id: str, env_id: str) -> list[str]:
        errors_list = []

        for local_role in self.roles:
            try:
                get_response = await self.roles_service.Get(
                    roles_pb2.GetRequest(
                        space_id=space_id,
                        role_id=local_role.id
                    )
                )

                role = get_response.role
            except grpc.RpcError as e:
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось получить роль {local_role.id}, {e.details()}")
                    continue

                role = None

            cloned_role = copy.deepcopy(local_role)
            cloned_role.space_id = space_id

            if role:
                if not role.environments:
                    role.environments[:] = [env_id]
                elif env_id not in role.environments:
                    role.environments.append(env_id)

                cloned_role.environments[:] = role.environments

                # Произвести мерж правил доступа
                for exist_rule in role.rules:
                    was_found = False

                    for local_rule in cloned_role.rules:
                        # Если запись с правами доступа на коллекцию найдена - правила нужно совместить
                        if local_rule.collection_id == exist_rule.collection_id:
                            was_found = True

                            local_rule.actions[:] = list(set(list(local_rule.actions) + list(exist_rule.actions)))

                            break

                    # Если правило для коллекций не было найдено - его нужно добавить
                    if not was_found:
                        cloned_role.rules.append(common_pb2.Rule(
                            collection_id=exist_rule.collection_id,
                            actions=list(exist_rule.actions)
                        ))

                try:
                    await self.roles_service.Update(
                        roles_pb2.UpdateRequest(
                            role=cloned_role
                        )
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось обновить роль {local_role.id}, {e.details()}")
            else:
                if env_id not in cloned_role.environments:
                    cloned_role.environments.append(env_id)

                try:
                    cloned_role = copy.deepcopy(local_role)
                    cloned_role.space_id = space_id

                    response = await self.roles_service.Create(
                        roles_pb2.CreateRequest(
                            role=cloned_role
                        ),
                    )
                except grpc.RpcError as e:
                    # На этапе install считается что ролей __нет__. При install с указанием force роли предварительно
                    # удаляются
                    errors_list.append(f"Не удалось создать роль {local_role.id}, {e.details()}")

        return errors_list

    # Работа с клиентами
    async def __check_clients(self, space_id: str) -> list[str]:
        errors_list = []

        for client in self.clients:
            try:
                await self.clients_service.Get(
                    clients_pb2.GetRequest(space_id=space_id, id=client.id)
                )
            except grpc.RpcError as e:
                errors_list.append(f"Не удалось получить клиент {client.id}, ошибка {e.details()}")

        return errors_list

    async def __remove_clients(self, space_id: str) -> list[str]:
        errors_list: list[str] = []

        for client in self.clients:
            try:
                await self.clients_service.Delete(
                    clients_pb2.DeleteRequest(
                        space_id=space_id, id=client.id
                    )
                )
            except grpc.RpcError as e:
                # Отсутствие клиента ошибкой не считается
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось удалить клиент {client.id}, {e.details()}")

        return errors_list

    async def __update_clients(self, space_id: str) -> list[str]:
        errors_list = []

        for local_client in self.clients:
            try:
                # Перед обновлением клиента предварительно нужно получить текущую запись чтобы скопировать оттуда
                # токены. Иначе после обновления расширения перестанут работать все приложения которые использовали
                # токен клиента
                get_response = await self.clients_service.Get(
                    clients_pb2.GetRequest(
                        space_id=space_id,
                        id=local_client.id
                    )
                )

                client = get_response.client
            except grpc.RpcError as e:
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось получить клиент {local_client.id}, {e.details()}")
                    continue

                client = None

            if client:
                try:
                    # Нужно чтобы у клиента каждый раз не слетали данные токенов при переустановке
                    # свойства oauth, api_key и tls должны браться из __созданного__ клиента
                    new_client = clients_pb2.Client(
                        id=local_client.id,
                        space_id=space_id,
                        name=local_client.name,
                        description=local_client.description,
                        disabled=client.disabled,
                        role_id=local_client.role_id,
                        oauth=client.oauth or local_client.oauth,
                        api_key=client.api_key or local_client.api_key,
                        tls=client.tls
                    )

                    await self.clients_service.Update(
                        clients_pb2.UpdateRequest(
                            client=new_client
                        )
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось обновить клиент {local_client.id}, {e.details()}")
            else:
                try:
                    cloned_client = copy.deepcopy(local_client)
                    cloned_client.space_id = space_id

                    await self.clients_service.Create(
                        clients_pb2.CreateRequest(
                            client=cloned_client
                        ),
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось создать клиент {local_client.id}, {e.details()}")

        return errors_list

    # Работа с коллекциями
    async def __remove_collections(self, space_id: str, env_id: str) -> list[str]:
        errors_list: list[str] = []

        for collection in self.collections:
            try:
                await self.collections_service.Delete(
                    collections_pb2.DeleteRequest(
                        space_id=space_id, env_id=env_id, collection_id=collection.id
                    )
                )
            except grpc.RpcError as e:
                # Отсутствие коллекции это не ошибка
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось удалить коллекцию {collection.id}, {e.details()}")

        return errors_list

    async def __check_collections(self, space_id: str, env_id: str) -> list[str]:
        errors_list = []

        for collection in self.collections:
            try:
                await self.collections_service.Get(
                    collections_pb2.GetRequest(space_id=space_id, env_id=env_id, collection_id=collection.id)
                )
            except grpc.RpcError as e:
                errors_list.append(f"Не удалось получить коллекцию {collection.id}, ошибка {e.details()}")

        return errors_list

    async def __update_collections(self, space_id: str, env_id: str) -> list[str]:
        """
            Метод __обновления__ коллекций. Миграция окружения требуется
            только в случае если одна или несколько схем коллекций изменялись. Алгоритм работы:
            1. Получить фактически существующую коллекцию из БД
            2. Обновить её в perxis
            3. Сравнить схему коллекции в расширении и в perxis
            4. Если схема изменена - обновить схему в perxis
            5. Если обновлялась хотя бы одна схема в perxis - запустить миграцию окружения
        """

        need_to_migrate_environment = False
        errors_list = []

        for local_collection in self.collections:
            try:
                # Необходимо получить текущую версию коллекции для того чтобы сравнить схемы
                get_response = await self.collections_service.Get(
                    collections_pb2.GetRequest(space_id=space_id, env_id=env_id, collection_id=local_collection.id)
                )

                collection = get_response.collection
            except grpc.RpcError as e:
                collection = None

            cloned_collection = copy.deepcopy(local_collection)
            cloned_collection.space_id = space_id
            cloned_collection.env_id = env_id

            # Коллекция может быть не найдена в случае если она добавлена в новой версии
            if not collection:
                try:
                    create_response = await self.collections_service.Create(
                        collections_pb2.CreateRequest(collection=cloned_collection)
                    )

                    collection = create_response.created
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось создать коллекцию {local_collection.id}, {e.details()}")

                    continue
            else:
                # если view-коллекция то не устанавливаем схему
                if MessageToDict(collection).get("view"):
                    continue
                try:
                    cloned_collection = copy.deepcopy(local_collection)
                    cloned_collection.space_id = space_id
                    cloned_collection.env_id = env_id

                    await self.collections_service.Update(
                        collections_pb2.UpdateRequest(collection=cloned_collection)
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось обновить коллекцию {local_collection.id}, {e.details()}")

                    continue

            diff = DeepDiff(
                json.loads(collection.schema or "{}"),
                json.loads(local_collection.schema or "{}"),
                ignore_numeric_type_changes=True,
                exclude_paths=["root['loaded']"]
            )
            if diff:
                need_to_migrate_environment = True

                set_schema_error_message = await self.__set_collection_schema(
                    space_id, env_id, local_collection.id, local_collection.schema
                )
                if set_schema_error_message:
                    errors_list.append(set_schema_error_message)

        if need_to_migrate_environment:
            migrate_environment_error_message = await self.__migrate_environment(space_id, env_id)
            if migrate_environment_error_message:
                errors_list.append(migrate_environment_error_message)

        return errors_list

    async def __migrate_environment(self, space_id: str, env_id: str) -> typing.Optional[str]:
        # Так как perxis может не сразу выставить коллекции / окружению статус ready операцию необходимо выполнять
        # с попытками

        attempt = 0
        is_ok = False
        error_message = None

        while attempt <= self.__max_attempts_count and not is_ok:
            time.sleep(self.__sleep_time)

            try:
                await self.environments_service.Migrate(environments_pb2.MigrateRequest(
                    space_id=space_id,
                    env_id=env_id,
                    options=environments_pb2.MigrateOptions(
                        wait=True
                    )
                ))

                is_ok = True
            except grpc.RpcError as e:
                # Если не удалось мигрировать окружение по любой причине кроме подготовки - это ошибка
                if "is preparing" not in e.details():
                    error_message = e.details()

                    # Для принудительного выхода из цикла
                    attempt = self.__max_attempts_count

            attempt += 1

        return error_message

    async def __set_collection_schema(self, space_id: str, env_id: str, collection_id: str, schema: str) -> typing.Optional[str]:
        # Так как perxis может не сразу выставить коллекции / окружению статус ready операцию необходимо выполнять
        # с попытками

        attempt = 0
        is_ok = False
        error_message = None

        while attempt <= self.__max_attempts_count and not is_ok:
            time.sleep(self.__sleep_time)

            try:
                await self.collections_service.SetSchema(
                    collections_pb2.SetSchemaRequest(
                        space_id=space_id,
                        env_id=env_id,
                        collection_id=collection_id,
                        schema=schema
                    )
                )

                is_ok = True
            except grpc.RpcError as e:
                # Если не удалось установить схему по любой причине кроме подготовки - это ошибка
                if "is preparing" not in e.details():
                    error_message = e.details()

                    # Для принудительного выхода из цикла
                    attempt = self.__max_attempts_count

            attempt += 1

        return error_message

    # Работа с действиями
    async def __check_actions(self, space_id: str, env_id: str) -> list[str]:
        errors_list = []

        ids = [data["id"] for data in self.actions if data.get("id")]

        items = []

        try:
            result = await self.items_service.Find(
                items_pb2.FindRequest(
                    space_id=space_id,
                    env_id=env_id,
                    collection_id=ACTIONS_COLLECTION_ID,
                    filter=items_pb2.Filter(q=[f"id in {ids}"]),
                    options=items_pb2.FindOptions(
                        options=common_pb2.FindOptions(
                            page_num=0, page_size=len(ids)
                        )
                    ),
                )
            )

            items = result.items
        except grpc.RpcError as e:
            errors_list.append(f"Не удалось получить данные о действиях, ошибка {e.details()}")
        finally:
            found_ids = [item.id for item in items]

            for action_id in found_ids:
                if action_id not in ids:
                    errors_list.append(f"Действие {action_id} не найдено")

        return errors_list

    async def __update_actions(self, space_id: str, env_id: str) -> list[str]:
        errors_list = []
        not_found = False

        for action in self.actions:
            action_item = make_action_item(space_id, env_id, action)

            try:
                await self.items_service.Update(
                    items_pb2.UpdateRequest(
                        item=items_pb2.Item(
                            id=action_item.id,
                            space_id=space_id,
                            env_id=env_id,
                            collection_id=ACTIONS_COLLECTION_ID,
                            data=action_item.data,
                        )
                    )
                )
            except grpc.RpcError as e:
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось обновить действие {action_item.id}, {e.details()}")
                    continue

                not_found = True
            except Exception as e:
                errors_list.append(f"Не удалось обновить действие {action_item.id}, {e}")

                continue

            if not_found:
                try:
                    await self.items_service.Create(
                        items_pb2.CreateRequest(
                            item=action_item
                        )
                    )
                except grpc.RpcError as e:
                    errors_list.append(f"Не удалось создать действие {action_item.id}, {e.details()}")
                except Exception as e:
                    errors_list.append(f"Не удалось создать действие {action_item.id}, {e}")

                    continue

            try:
                await self.items_service.Publish(
                    items_pb2.PublishRequest(
                        item=action_item
                    )
                )
            except grpc.RpcError as e:
                # Если action уже опубликован
                if "item cannot be published in this state" not in e.details():
                    errors_list.append(f"Не удалось опубликовать действие {action_item.id}, {e.details()}")
            except Exception as e:
                errors_list.append(f"Не удалось опубликовать действие {action_item.id}, {e}")

        return errors_list

    async def __remove_actions(self, space_id: str, env_id: str) -> list[str]:
        errors_list = []
        for action in self.actions:
            action_item = make_action_item(space_id, env_id, action)

            try:
                await self.items_service.Delete(
                    items_pb2.DeleteRequest(
                        item=items_pb2.Item(
                            id=action_item.id,
                            space_id=space_id,
                            env_id=env_id,
                            collection_id=action_item.collection_id,
                        ),
                        options=items_pb2.DeleteOptions(
                            update_attrs=True,
                            erase=True
                        )
                    )
                )
            except grpc.RpcError as e:
                # Отсутствие действия это не ошибка
                if "not found" not in e.details():
                    errors_list.append(f"Не удалось удалить действие {action.get('id', 'n/a')}, {e.details()}")
            except Exception as e:
                errors_list.append(f"Не удалось удалить действие - {e}")

        return errors_list

    async def __check_items(self, space_id: str, env_id: str) -> list[str]:
        errors_list: list[str] = []
        wrapper = PerxisItemsWrapper(
            self.items_service,
        )

        for item in self.items:
            all_rules_satisfied = await item.all_rules_is_satisfied(space_id, env_id)

            if not all_rules_satisfied:
                continue

            try:
                message = await wrapper.find(
                    collection_id=item.collection_id,
                    env_id=env_id,
                    space_id=space_id,
                    limit=1,
                    offset=0,
                    fields=["id"],
                    filters=[f"{item.identifier_field} == '{item.identifier}'"]
                )

                if not message.total:
                    errors_list.append(f"Item {item.identifier} не найден")
            except Exception as e:
                errors_list.append(
                    f"Не удалось проверить item {item.identifier} "
                    f"коллекции {item.collection_id} - {e}"
                )

        return errors_list

    async def __update_items(self, space_id: str, env_id: str) -> list[str]:
        errors_list: list[str] = []
        wrapper = PerxisItemsWrapper(
            self.items_service,
        )

        for item in self.items:
            all_rules_satisfied = await item.all_rules_is_satisfied(space_id, env_id)

            if not all_rules_satisfied:
                continue

            try:
                message = await wrapper.find(
                    collection_id=item.collection_id,
                    env_id=env_id,
                    space_id=space_id,
                    limit=1,
                    offset=0,
                    filters=[f"{item.identifier_field} == '{item.identifier}'"]
                )

                if message.items:
                    item_in_perxis = message.items[0]
                else:
                    item_in_perxis = None
            except Exception as e:
                if hasattr(e, "details") and "not found" in e.details():
                    is_error = False
                else:
                    is_error = True

                if is_error:
                    errors_list.append(
                        f"Не удалось получить item {item.identifier} "
                        f"коллекции {item.collection_id} - {e}"
                    )

                continue

            try:
                if item_in_perxis:
                    # Если установлен запрет на изменение item'ов
                    if not item.with_update:
                        continue

                    # Для того чтобы не затереть изменения в perxis
                    # Нужно смержить данные. Логика работы:
                    # 1. Данные которые указаны в `data` в расширении - в приоритете, они замещают то что в perxis
                    # 2. Данные которые есть в perxis но нет в расширении - дополняются
                    await wrapper.update(
                        collection_id=item.collection_id,
                        item_id=item_in_perxis.id,
                        space_id=space_id,
                        env_id=env_id,
                        data=item.merge_data(MessageToDict(item_in_perxis)["data"]),
                    )
                else:
                    message = await wrapper.create(
                        collection_id=item.collection_id,
                        space_id=space_id,
                        env_id=env_id,
                        data=item.struct,
                    )

                    item_in_perxis = message.created

                await wrapper.publish(
                    item_id=item_in_perxis.id,
                    collection_id=item.collection_id,
                    space_id=space_id,
                    env_id=env_id
                )
            except Exception as e:
                errors_list.append(
                    f"Не удалось записать item {item.identifier} "
                    f"коллекции {item.collection_id} - {e}"
                )

                continue

        return errors_list

    async def __remove_items(self, space_id: str, env_id: str) -> list[str]:
        errors_list: list[str] = []
        wrapper = PerxisItemsWrapper(
            self.items_service,
        )

        for item in self.items:
            all_rules_satisfied = await item.all_rules_is_satisfied(space_id, env_id)

            if not all_rules_satisfied:
                continue

            # Если установлен запрет на удаление item'ов
            if not item.with_delete:
                continue

            try:
                message = await wrapper.find(
                    collection_id=item.collection_id,
                    env_id=env_id,
                    space_id=space_id,
                    limit=1,
                    offset=0,
                    fields=["id"],
                    filters=[f"{item.identifier_field} == '{item.identifier}'"]
                )

                if message.items:
                    await wrapper.delete(
                        item_id=message.items[0].id,
                        collection_id=message.items[0].collection_id,
                        space_id=space_id,
                        env_id=env_id,
                    )
            except Exception as e:
                if hasattr(e, "details") and "not found" in e.details():
                    is_error = False
                else:
                    is_error = True

                if is_error:
                    errors_list.append(
                        f"Не удалось удалить item {item.identifier} "
                        f"коллекции {item.collection_id} - {e}"
                    )

        return errors_list

    async def __update_view_role(self, space_id: str, env_id: str, mode: str = "add") -> list[str]:
        errors = []

        # Нужны только относящиеся к синхронизации элементы
        items_for_view_role = [
            item
            for item
            in self.items
            if (isinstance(item, SyncPolicyItem) or item.collection_id == "hoop_item_sync_policies")
            and item.data["export_view"]
            and await item.all_rules_is_satisfied(space_id=space_id, env_id=env_id)
        ]

        if not items_for_view_role:
            return errors

        try:
            message = await self.roles_service.Get(
                roles_pb2.GetRequest(
                    space_id=space_id,
                    role_id="view"
                )
            )

            role = message.role
        except grpc.RpcError as e:
            if "not found" not in e.details():
                errors.append(f"Не удалось получить роль view, {e.details()}")

            role = None

        if not role:
            try:
                message = await self.roles_service.Create(
                    roles_pb2.CreateRequest(
                        role=roles_pb2.Role(
                            id="view",
                            space_id=space_id,
                            description="Роль для view коллекций",
                            rules=[],
                            environments=["*"],
                            allow_management=False,
                        )
                    ),
                )

                role = message.created
            except grpc.RpcError as e:
                errors.append(f"Не удалось создать роль view, {e.details()}")

                return errors

        # Произвести мерж правил доступа
        for item in items_for_view_role:
            actions_from_item = []

            if not item.data["deny_read"]:
                actions_from_item.append(common_pb2.READ)

            if not item.data["deny_create"]:
                actions_from_item.append(common_pb2.CREATE)

            if not item.data["deny_publish"]:
                actions_from_item.append(common_pb2.UPDATE)

            if not item.data["deny_delete"]:
                actions_from_item.append(common_pb2.DELETE)

            rule_was_found = False
            for rule in role.rules:
                if rule.collection_id != item.data["collection"]:
                    continue

                rule_was_found = True

                if mode == "add":
                    modified_actions = list(set(list(rule.actions) + actions_from_item))
                else:
                    modified_actions = [
                        action
                        for action
                        in rule.actions
                        if action not in actions_from_item
                    ]

                rule.actions[:] = modified_actions

            # Если правила для коллекции нет и при этом доступы нужно __добавлять__
            if not rule_was_found and mode == "add":
                role.rules.append(
                    common_pb2.Rule(
                        collection_id=item.data["collection"],
                        actions=actions_from_item,
                    )
                )

        try:
            await self.roles_service.Update(
                roles_pb2.UpdateRequest(
                    role=role
                )
            )
        except grpc.RpcError as e:
            errors.append(f"Не удалось обновить роль view, {e.details()}")

        return errors


    async def install(self, space_id: str, env_id: str, use_force: bool) -> list[str]:
        errors = []

        try:
            errors += await self.__update_collections(space_id, env_id)
            errors += await self.__update_roles(space_id, env_id)
            errors += await self.__update_clients(space_id)
            errors += await self.__update_actions(space_id, env_id)
            errors += await self.__update_items(space_id, env_id)
            errors += await self.__update_view_role(space_id, env_id)
        except Exception as e:
            logger.exception(e)
            errors.append(f"Во время установки было необработанное исключение - {e}")

        return errors

    async def update(self, space_id: str, env_id: str, use_force: bool) -> list[str]:
        errors = []

        try:
            errors += await self.__update_collections(space_id, env_id)
            errors += await self.__update_roles(space_id, env_id)
            errors += await self.__update_clients(space_id)
            errors += await self.__update_actions(space_id, env_id)
            errors += await self.__update_items(space_id, env_id)
            errors += await self.__update_view_role(space_id, env_id)
        except Exception as e:
            logger.exception(e)
            errors.append(f"Во время обновления было необработанное исключение - {e}")

        return errors

    async def check(self, space_id: str, env_id: str) -> list[str]:
        errors = []

        try:
            errors += await self.__check_collections(space_id, env_id)
            errors += await self.__check_roles(space_id)
            errors += await self.__check_clients(space_id)
            errors += await self.__check_actions(space_id, env_id)
            errors += await self.__check_items(space_id, env_id)
        except Exception as e:
            logger.exception(e)
            errors.append(f"Во время проверки было необработанное исключение - {e}")

        return errors

    async def uninstall(self, space_id: str, env_id: str, use_remove: bool) -> list[str]:
        errors = []
        if use_remove:
            try:
                errors += await self.__remove_collections(space_id, env_id)
                errors += await self.__remove_clients(space_id)
                errors += await self.__remove_roles(space_id)
                errors += await self.__remove_actions(space_id, env_id)
                errors += await self.__remove_items(space_id, env_id)
                errors += await self.__update_view_role(space_id, env_id, "remove")
            except Exception as e:
                logger.exception(e)
                errors.append(f"Во время удаления данных было необработанное исключение - {e}")

        return errors
