import abc
import typing
import logging

from perxis.collections import collections_pb2
from perxis.extensions import manager_service_pb2


if typing.TYPE_CHECKING:
    from perxis.extensions.item_models import AbstractItem


class AbstractRule(metaclass=abc.ABCMeta):
    """
        Абстрактное правило для обработки создания item'ов в системе. Правила могут
        использоваться в кейсах:
            1. В случае если item опционален и невозможность его создания не приведёт к невозможности
               использовать само расширение
            2. Если при создании item'а нужно обогатить его данные чем-то внешним, не относящимся к
               текущему расширению. Например - при создании item'ов для цепочки сервиса уведомлений.
               Каждый последующий item может с помощью правил получать результат создания предыдущего
               и на базе этих данных формировать ref

        Так как сервисов в ExtensionService очень много, для того чтобы код их проброса в объект rule
        не был излишне длинным - это делается неявным пробросом свойств из ExtensionService. Таким
        образом и код короче получается, и в случае если будут добавлены какие то новые сервисы они
        автоматически будут проброшены сюда.
    """

    @property
    def rule_name(self):
        return self.__class__.__name__

    @property
    def logger(self):
        return logging.getLogger(self.rule_name)

    def bind_services(self, service_list: list[tuple[str, object]]):
        """
            Нужно для автоматического подключения сервисов из ExtensionService. Проблема
            в том что их очень много и если руками явно указывать - будет большой boilerplate
        """

        for service_name, service in service_list:
            setattr(self, service_name, service)

    @abc.abstractmethod
    async def act(self, item: "AbstractItem", space_id: str, env_id: str):
        raise NotImplementedError()

    async def __call__(self, item: "AbstractItem", space_id: str, env_id: str) -> bool:
        try:
            await self.act(item, space_id, env_id)

            return True
        except Exception as e:
            self.logger.warning(
                f"Правило {self.rule_name} не "
                f"сработало для {item.collection_id} (item {item.identifier}), "
                f"ошибка - {e}"
            )

            return False


class IfCollectionExists(AbstractRule):
    """
        Правило для опциональных item'ов которые нужно создавать в случае наличия
        определённой коллекции в perxis. Может быть полезно если коллекция широко используется
        но в данный момент устанавливается только вручную и не привязана ни к одному из расширений
    """

    async def act(self, item: "AbstractItem", space_id: str, env_id: str):
        message = await self.collections_service.Get(
            collections_pb2.GetRequest(
                space_id=space_id,
                env_id=env_id,
                collection_id=item.collection_id
            )
        )

        if not message.collection:
            raise ValueError(f"Коллекция {item.collection_id} не найдена")


class IfExtensionInstalled(AbstractRule):
    """
        Правило для опциональных item'ов которые нужно создавать только в случае
        наличия определённого расширения в perxis
    """

    required_extension: str

    def __init__(self, required_extension: str):
        self.required_extension = required_extension

    async def act(self, item: "AbstractItem", space_id: str, env_id: str):
        registered_extensions = await self.ext_manager_service.ListRegisteredExtensions(
            manager_service_pb2.ListRegisteredExtensionsRequest()
        )

        ext_is_installed = False
        for ext in registered_extensions.extensions:
            if ext.extension == self.required_extension:
                ext_is_installed = True
                break

        if not ext_is_installed:
            raise ValueError(
                f"Расширение {self.required_extension} не установлено в perxis"
            )
